#include <stdio.h>
#include <stdlib.h>
#include <string.h>


typedef unsigned long dword;
typedef unsigned int word;
typedef unsigned char byte;


int     TYPE=0; // 48/128 mode
FILE    *File;
int     ErrRd;

byte    *page[8];
byte    SpecSeg[16];


/*   */
struct DatCPU {
    word RegPC; //  PC
    word RegSP; //  SP
    word RegIX; //  IX
    word RegIY; //  IY
    union {word RegAF;struct {byte RegF;byte RegA;};}; //  AF
    union {word RegBC;struct {byte RegC;byte RegB;};}; //  BC
    union {word RegDE;struct {byte RegE;byte RegD;};}; //  DE
    union {word RegHL;struct {byte RegL;byte RegH;};}; //  HL
    union {word RegAFa;struct {byte RegFa;byte RegAa;};}; //  AF'
    union {word RegBCa;struct {byte RegCa;byte RegBa;};}; //  BC'
    union {word RegDEa;struct {byte RegEa;byte RegDa;};}; //  DE'
    union {word RegHLa;struct {byte RegLa;byte RegHa;};}; //  HL'
	byte RegR; //  R
	byte RegI; //  I
    byte    Port7FFD;
    byte    IFF1,IFF2,IM;

} CPU;


// Headers

  //SNA

struct _Sna48Hdr
{
    byte    I;
    word    HLa,DEa,BCa,AFa;
    word    HL,DE,BC,IY,IX;
    byte    Int;
    byte    R;
    word    AF,SP;
    byte    IntMode;
    byte    Border;
} Sna48Hdr;

struct  _Sna128
{
    word    PC;
    byte    P7FFD;
    byte    u;
} Sna128;

 // Z80

struct _Z80_1
{
    byte    A;
    byte    F;
    word    BC;
    word    HL;
    word    PC;
    word    SP;
    byte    I;
    byte    R;

    byte    F0; //
 //    unsigned R_7:0x1;           //Bit 0  : Bit 7 of the R-register
 //    unsigned Border:0x3;        //Bit 1-3: Border colour
 //    unsigned SROM:0x1;          //1=Basic SamRom switched in
 //    unsigned CompFlag:0x1;      //1=Block of data is compressed

    word    DE;
    word    BCa;
    word    DEa;
    word    HLa;
    byte    Aa;
    byte    Fa;
    word    IY;
    word    IX;
    byte    IFF1;               // Interrupt flipflop, 0=DI, otherwise EI
    byte    IFF2;               // IFF2 (not particularly important...)
    byte    F1;
 //    byteIntMode:0x2;       // Bit 0-1: Interrupt mode (0, 1 or 2)
 //    unsigned I2E:0x1;           // Bit 2  : 1=Issue 2 emulation
 //    unsigned Dfrq:0x1;          // Bit 3  : 1=Double interrupt frequency
 //    unsigned Vsync:0x2;         // Bit 4-5: 1=High vid sync, 3=Low vid sync, 0,2=Normal
 //    unsigned JoyType:0x2;       //Bit 6-7: 0=Cursor/Protek/AGF joystick

} Z80_H1;

struct _Z80_2
{
    word    Add_Len;    //Len of add hdr (23-V2.01)(54-V3.0)
    word    PC;         //Program counter
    byte    HardMode;   //Hardware mode
 //:0 - 48k
 //:1 - 48k + If.1
 //:2 - SamRam
 //:3 - 128k
 //:4 -  128k + If.1

    byte    P7FFD;      //port 7ffd
    byte    I1;         //Contains 0FF if Interface I rom paged
    byte    FL2;
 //:00000001  if R register emulation on,
 //:00000010  1 if LDIR emulation on
    byte    RegIndex;   //Last OUT to fffd (soundchip register number)
    byte    AY[16];     //Contents of the sound chip registers
} Z80_H2;

struct _Z80_3
{
    word   Lt;//       Low T state counter
    byte   Ht;//       Hi T state counter
    byte   QL;//       Flag byte used by Spectator (QL spec. emulator), Ignored by Z80 when loading, zero when saving
    byte   MGTR;//     0FF if MGT Rom paged
    byte   MR;//       0FF if Multiface Rom paged. Should always be 0.
    byte   Ram0;//     0FF if 0-8191 is RAM
    byte   Ram1;//     0FF if 8192-16383 is RAM
    byte   Key0[10];// 5x keyboard mappings for user defined joystick
    byte   Key1[10];// 5x ascii word: keys corresponding to mappings above
    byte   Mgtt;//     MGT type: 0=Disciple+Epson,1=Discipls+HP,16=Plus D
    byte   Dsc1;//     Disciple inhibit button status: 0=out, 0ff=in
    byte   Dsc2;//     Disciple inhibit flag: 0=rom pageable, 0ff=not
 //    byte   unk; //     ?????????
} Z80_H3;

struct _CompHdr
{
    word    Len;
    byte    Page;
} CmpHdr;


// End HEADERS

// Z80 Loaders ...

byte GetByte(void)
{

    byte    buf[20000];
    static    word    bptr=16384;

    if (  bptr == 16384  )
    {
       fread( buf,1,16384,File);
       bptr=0;
    }
    return buf[bptr++];
}

void PutByte(word addr,byte b)
{
    (page[SpecSeg[addr>>0xE]])[addr&0x3FFF]=b;
}



void Z80Read48( word dst)
{
    byte    last=255,c1,c2,len;
    int     curr;
    word    ccc=0;

    for(;;)
    {
        c1=GetByte();
        if ( c1 == 0xed )
        {
            c2=GetByte();
            if ( c2 == 0xed)
            {
                len=GetByte();
                if ( ( len == 0 ) && ( last == 0 ) ) break;
                c1=GetByte();
                len--;

                while(len)
                {
                    PutByte(dst++,c1);
                    len--;
                    ccc++;
                }
            }
            else
            {
                PutByte(dst++,c1);
                ccc++;
                c1=c2;
            }
        }
        PutByte(dst++,c1);
        ccc++;
        last=c1;
    }
}


void Z80Read128(byte *dst,int ImgLen)
{
    static byte    buf[17000];
    byte    *src=buf;
    byte    *end;

    byte    *TmpDST=dst;

    byte    len;

    end=src+ImgLen;

    if ( fread( buf,1,ImgLen, File) != ImgLen ) { ErrRd=1; return; };

    byte    c1,c2;

    while ( src < end )
    {
        c1=*src++;
        if ( c1 == 0xed )
        {
            c2=*src++;
            if ( c2 == 0xed)
            {
                len=*src++;
                c1=*src++;
                len--;

                while(len)
                {
                    *dst++=c1;
                    len--;
                }
            }
            else
            {
                *dst++=c1;
                c1=c2;
            }
        } // C1 == 0xed
        *dst++=c1;
    } // While Src<End
    printf(" ExpLen=%d ",TmpDST-dst-1);

}


void ReadZ80(char *FileName)
{
    byte version=0;
    char Cver[3][5]={"1.45","2.01","3.0"};
 //                    0  1  2  3  4  5  6  7  8  9 10 11
    byte map[2][12]={{10,10,10,10, 1, 2,10,10, 5,10,10,10}, // 48K Mode
                     {10,10,10, 0, 1, 2, 3, 4, 5, 6, 7,10}};    // 128k Mode

    byte HType[2][7]={{0,0,0,1,1,0,0},{0,0,0,1,1,1,1}}; // 0 48/ 1 128


    if ( ( File=fopen(FileName,"rb") ) >0 )
    {

        if ( fread( &Z80_H1, 1, sizeof(Z80_H1), File) != sizeof(Z80_H1) ) ErrRd=1;

        SpecSeg[1]=5; // RAM PAGEs
        SpecSeg[2]=1;
        SpecSeg[3]=2;

        CPU.RegPC=Z80_H1.PC;// 0x4a6;//
        CPU.RegSP=Z80_H1.SP;
        CPU.RegIX=Z80_H1.IX;
        CPU.RegIY=Z80_H1.IY;
        CPU.RegAF=Z80_H1.F | Z80_H1.A << 8;
        CPU.RegBC=Z80_H1.BC;
        CPU.RegDE=Z80_H1.DE;
        CPU.RegHL=Z80_H1.HL;
        CPU.RegAFa=Z80_H1.Fa | Z80_H1.Aa << 8;
        CPU.RegBCa=Z80_H1.BCa;
        CPU.RegDEa=Z80_H1.DEa;
        CPU.RegHLa=Z80_H1.HLa;
        CPU.RegR=Z80_H1.R | (Z80_H1.F0 & 1)<<7;
        CPU.RegI=Z80_H1.I;

        CPU.IFF1=(Z80_H1.IFF1)?4:0;
        CPU.IFF2=(Z80_H1.IFF2)?4:0;
        CPU.IM=Z80_H1.F1 & 3; //IntMode;


        CPU.Port7FFD=0x30;      // 48k Mode

        if ( Z80_H1.PC == 0 )
        {
            version=1;
            if ( fread( &Z80_H2, 1,sizeof(Z80_H2), File ) != sizeof(Z80_H2) ) ErrRd=1;
            if ( Z80_H2.Add_Len != 23 )
            {
                if (Z80_H2.Add_Len != 54)  ErrRd = 1;
                if ( fread( &Z80_H3, 1, sizeof(Z80_H3), File) != sizeof(Z80_H3) ) ErrRd=1;
                version=2;
            }
        }

        if ( version == 0 )
        {                                       // 48 k only
            if (Z80_H1.F0 && 0x10 /*CompFlag*/ )
            {
                Z80Read48( 0x4000 );
            } // compress
            else
            {
                if ( fread( page[SpecSeg[1]] , 1, 16384 , File) != 16384 ) ErrRd=1;
                if ( fread( page[SpecSeg[2]] , 1, 16384 , File) != 16384 ) ErrRd=1;
                if ( fread( page[SpecSeg[3]] , 1, 16384 , File) != 16384 ) ErrRd=1;
            } // no compress

        } // version 0
        else
        {                               // 48 & 128 k
            int mod=HType[(version>0)?1:0][Z80_H2.HardMode]; // 0 - 48, 1 -128

            while (  fread( &CmpHdr , 1,sizeof(CmpHdr) , File ) == sizeof(CmpHdr) )
            {
                printf(" Page=%d,Len=%d",CmpHdr.Page,CmpHdr.Len);
                Z80Read128(page[ map[mod][CmpHdr.Page] ] , CmpHdr.Len);
                printf(" pl=%d \r\n",map[mod][CmpHdr.Page]);

            }
            if ( mod ) // 128k Mode
            {
                CPU.Port7FFD=Z80_H2.P7FFD;
            }
            CPU.RegPC=Z80_H2.PC;

        } // mode 1&2

        if ( ErrRd )
        {
            printf("訡 ⥭,   ० 48k.\r\n");
        }
        else
        {
            printf(".Z80 Version %s\r\n",Cver[version]);
        } // ErrRd

        fclose(File);

        TYPE = (CPU.Port7FFD == 0x30 )?48:128;
    }
    else
    {
        printf("訡  䠩 %s\r\n ",FileName);
    } // FileOpen Error
}

// SNA128 Save
void SaveSNA128(char *FileName)
{
    byte    pg[]={1,1,0,1,1,0,1,1};


    if ( ( File=fopen(FileName,"wb") ) >0 )
    {
        Sna128.PC   =CPU.RegPC;
        Sna48Hdr.SP =CPU.RegSP;
        Sna48Hdr.IX =CPU.RegIX;
        Sna48Hdr.IY =CPU.RegIY;
        Sna48Hdr.AF =CPU.RegAF;
        Sna48Hdr.BC =CPU.RegBC;
        Sna48Hdr.DE =CPU.RegDE;
        Sna48Hdr.HL =CPU.RegHL;
        Sna48Hdr.AFa=CPU.RegAFa;
        Sna48Hdr.BCa=CPU.RegBCa;
        Sna48Hdr.DEa=CPU.RegDEa;
        Sna48Hdr.HLa=CPU.RegHLa;
        Sna48Hdr.R  =CPU.RegR & 0x7f;
        Sna48Hdr.I  =CPU.RegI;

        Sna48Hdr.Int=CPU.IFF1>>2 | CPU.IFF2 >> 1;
        Sna48Hdr.IntMode=CPU.IM;
        Sna48Hdr.Border=0;

        Sna128.P7FFD=CPU.Port7FFD;
        Sna128.u=0;

        fwrite(&Sna48Hdr,sizeof(Sna48Hdr),1,File);

        fwrite(page[5],16384,1,File);
        fwrite(page[2],16384,1,File);

        pg[CPU.Port7FFD & 7]=0;
        fwrite(page[CPU.Port7FFD & 7],16384,1,File);

        fwrite(&Sna128,sizeof(Sna128),1,File);

        for ( int i=0;i<8;i++ )
            if (pg[i]) fwrite(page[i],16384,1,File);

        fclose(File);
        printf("Sna 128 write ok.\r\n");

    }
    else
    {
        printf(" Error open file: %s\r\n",FileName);
    }
}

void SaveZ80Page(int Page)
{
    static byte dst[30000];
    byte    *src=page[Page];

    word    S_Len=16384;
    word    D_Len=0;
    word    S_Ptr=0;
    word    D_Ptr=0;
    word    sp2,seqlen;
    int     LastED=0;

    while ( S_Len )
    {
        if ( !LastED )
        {

            if ( src[S_Ptr]==src[S_Ptr+1] )
            {
                sp2=S_Ptr+1;
                seqlen=1;
                while ( (src[S_Ptr]==src[sp2++]) &&
                        (seqlen<255) &&
                        ((S_Len-seqlen)>0) )            seqlen++;

                if ( (seqlen>4) || (src[S_Ptr] == 0xed) )
                {
                    dst[D_Ptr++]=0xed       ;D_Len++;
                    dst[D_Ptr++]=0xed       ;D_Len++;
                    dst[D_Ptr++]=seqlen     ;D_Len++;
                    dst[D_Ptr++]=src[S_Ptr] ;D_Len++;
                    S_Len-=seqlen;
                    S_Ptr+=seqlen;
                    LastED=2;
                }
            }
        }
        if ( LastED == 2) { LastED=0; }
        else
        {
            LastED=(src[S_Ptr]==0xed)?!0:0;
            dst[D_Ptr++]=src[S_Ptr++];
            S_Len--;D_Len++;
        }
    }

    CmpHdr.Page=Page+3;
    CmpHdr.Len=D_Len;

    fwrite(&CmpHdr,sizeof(CmpHdr),1,File);

    fwrite(dst,D_Len,1,File);
}

void SaveZ80(char *FileName,int M128,int V3Flag)
{
    byte  pg[2][8]={{0,1,1,0,0,1,0,0},
                    {1,1,1,1,1,1,1,1}};

    if ( ( File=fopen(FileName,"wb") ) >0 )
    {

        Z80_H1.PC=0;

        Z80_H1.SP=CPU.RegSP;
        Z80_H1.IX=CPU.RegIX;
        Z80_H1.IY=CPU.RegIY;
        Z80_H1.F=CPU.RegAF & 0xff;
        Z80_H1.A=CPU.RegAF >>8;
        Z80_H1.BC=CPU.RegBC;
        Z80_H1.DE=CPU.RegDE;
        Z80_H1.HL=CPU.RegHL;

        Z80_H1.Fa=CPU.RegAFa & 0xff;
        Z80_H1.Aa=CPU.RegAFa >>8;
        Z80_H1.BCa=CPU.RegBCa;
        Z80_H1.DEa=CPU.RegDEa;
        Z80_H1.HLa=CPU.RegHLa;

        Z80_H1.R=CPU.RegR & 0x7f | (Z80_H1.F0 & 1)<<7;
        Z80_H1.F0=CPU.RegR >>7 | 0x20; // Bit7 & compressed

        Z80_H1.I=CPU.RegI;

        Z80_H1.IFF1=CPU.IFF1?1:0;
        Z80_H1.IFF2=CPU.IFF2?1:0;
        Z80_H1.F1=CPU.IM | 0x40; //IntMode & kempston joystick;

        fwrite(&Z80_H1,sizeof(Z80_H1),1,File);

        Z80_H2.PC=CPU.RegPC;
        Z80_H2.Add_Len=V3Flag?54:23;
        Z80_H2.HardMode=V3Flag?4:3;   //Hardware mode

        if ( M128==0 )  Z80_H2.HardMode=0; // 48K mode

        Z80_H2.P7FFD=CPU.Port7FFD;      //port 7ffd
        Z80_H2.I1;         //Contains 0FF if Interface I rom paged
        Z80_H2.FL2=3;
        Z80_H2.RegIndex=0;   //Last OUT to fffd (soundchip register number)
        for (int i=0;i<17;i++) Z80_H2.AY[i]=0;     //Contents of the sound chip registers

        fwrite(&Z80_H2,sizeof(Z80_H2),1,File);


        if ( V3Flag ){
            Z80_H3.Lt=0;//       Low T state counter
            Z80_H3.Ht=0;//       Hi T state counter
            Z80_H3.QL=0;//       Flag byte used by Spectator (QL spec. emulator), Ignored by Z80 when loading, zero when saving
            Z80_H3.MGTR=0;//     0FF if MGT Rom paged
            Z80_H3.MR=0;//       0FF if Multiface Rom paged. Should always be 0.
            Z80_H3.Ram0=0;//     0FF if 0-8191 is RAM
            Z80_H3.Ram1=0;//     0FF if 8192-16383 is RAM
            Z80_H3.Mgtt=0;//     MGT type: 0=Disciple+Epson,1=Discipls+HP,16=Plus D
            Z80_H3.Dsc1=0;//     Disciple inhibit button status: 0=out, 0ff=in
            Z80_H3.Dsc2=0;//     Disciple inhibit flag: 0=rom pageable, 0ff=not

            fwrite(&Z80_H3,sizeof(Z80_H3),1,File);
        }

        for ( i=0;i<8;i++ ){
            if ( pg[M128][i] ){
                printf(" Page %d\r\n",i);
                SaveZ80Page(i);
            }
        }
    }
    else
    {
        printf(" Error open file: %s\r\n",FileName);
    }

}


void SaveRAW(char *FileName)
{
    char F1[80],F2[80];
    char    n[]=".0";
    char    *t;


    strcpy(F1,FileName);
    t=strrchr(F1,'.');
    if ( t != NULL )
    {
        *t=0;
    }


    strcpy(F2,F1);
    strcat(F2,".REG");
    File=fopen(F2,"wb");
    fwrite(&CPU,sizeof(CPU),1,File);
    fclose(File);

    for ( int i=0;i<8;i++ ){
        printf(" Page %d\r\n",i);

        n[1]=0x30+i;

        strcpy(F2,F1);
        strcat(F2,n);
        File=fopen(F2,"wb");
        fwrite(page[i],16384,1,File);
        fclose(File);
    }

}



int GetMem(void)
{
    for ( int i=0;i<8;i++ )
        if ( ( page[i] = (byte *)malloc(16*1024) )  == NULL )  return 0;
    return  1;
}

void FreeMem(void)
{
    for ( int i=0;i<8;i++ ) free( page[i] );
}

void main(int argc, char **argv)
{
    GetMem();
    if ( argc < 3 )
    {
        printf(" usage: %s name1 name2\r\n",argv[0]);
        return;
    }
    ReadZ80(argv[1]);
 //    SaveSNA128(argv[2]);
    SaveZ80(argv[2],0,0);
 //  SaveRAW(argv[2]);
    FreeMem();
}




